(:~
: This module creates the edition views.
: OLD DOCS: single edition view. According to the GET parameter
: "edition" this module gets from edition.xqm all common text lines and the ones
: that fit a tei:rdg are selected.
:
: The whole module is called by xml-utils.xqm.
:
: @version 1.0
: @since 7.0.0
: @author Simon Sendler
: @author Hannes Riebl
: @author Michelle Weidling
:)

xquery version "3.1";

module namespace bdn-transform="http://bdn-edition.de/ns/bdn-transform";

declare namespace bdn="http://www.bdn-edition.de/ns/1.0";
declare namespace tei="http://www.tei-c.org/ns/1.0";
declare default element namespace "http://www.w3.org/1999/xhtml";

import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace id-file="https://bdn-edition.de/ns/id-file" at "id-file.xqm";
import module namespace whitelist="https://bdn-edition.de/ns/whitelist" at "whitelist.xqm";

declare variable $bdn-transform:id-map :=
    map:merge
    (
        for $id in id-file:get-id-file()//bdn:id
            return map:entry($id/string(), $id/parent::bdn:file/@uri/string())
    );
declare variable $bdn-transform:colors :=
    map {
        "a": "#359d93",
        "ab":  "#533380",
        "b":  "#823066",
        "c":  "#df6b6a",
        "d":  "#f4b771",
        "e":  "#abd1e8",
        "f":  "#43640b",
        "z":  "#afca54",
        "bc": "#ded96f",
        "bd": "#3e5e04",
        "other": "#a4bfd1",
        "quote": "#79a6d2"
    };

declare function bdn-transform:transform(
    $nodes as node()*, $filename as xs:string, $mode as xs:string)
as node()*
{
    for $node in $nodes
    return
        try {
            (: this is obscure. why is a span with the id of the current node created before its transformation?
             : saved for remediation with [issue 321](https://gitlab.gwdg.de/bibliothek-der-neologie/bdn-SADE/-/issues/321)
             : => this is a candidate for removal
             :)
            if ( $node/@xml:id ) then
                element span
                {
                    attribute class { "text-anchor" },
                    attribute id { $node/@xml:id }
                }
            else
                (),
            if ( $node/@copyOf ) then
                let $content := bdn-transform:copy-node($node, $filename, $mode)
                return
                    local:main($content, $filename, $mode)
            else
                local:main($node, $filename, $mode),
                bdn-transform:make-whitespace-after($node)
        }
        catch * {
            bdn-transform:error-message($err:code, $err:description, $node, "bdn-transform:transform", $filename)
        }
};

declare function local:main(
    $node as node()*, $filename as xs:string, $mode as xs:string)
as node()*
{
    typeswitch( $node )
        case text()
        return
            text { replace($node, "@", " ") }
        case comment()
            return ()
        case element ( tei:ab )
            return
                if ( $node[@type = "half-title"] ) then
                    element div
                    {
                        attribute class { $node/@type },
                        bdn-transform:transform($node/node(), $filename, $mode)
                    }
                else
                    bdn-transform:transform($node/node(), $filename, $mode)
        case element ( tei:abbr )
        return
            if ( $mode = "digital" ) then
                let $hash-input := serialize($node) || util:random()
                let $id := util:hash(generate-id($node), "md5")
                let $ref := "#" || $id
                return
                (
                    element span
                    {
                        attribute class { "expan-content" },
                        attribute href { $ref },
                        attribute data-toggle { "modal" },
                        bdn-transform:transform($node/node(), $filename, $mode),
                        bdn-transform:make-modal($node, $mode, $id, $filename)
                    }
                )
            else
                bdn-transform:transform($node/node(), $filename, $mode)
        case element ( tei:aligned )
        return
            element span
            {
                attribute class {$node/@rend},
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:app )
        return
            bdn-transform:apparatus($node, $filename, $mode)
        case element ( tei:back )
        return
            bdn-transform:transform($node/node(), $filename, $mode)
        case element ( tei:cb )
        return
            element span
            {
                attribute class {"cb"},
                string( "|" || replace( $node/@edRef, '[# ]+', '' ) || $node/@n || "|" )
            }
        case element ( tei:cell )
        return
            element td
            {
                if ( $node/@role = "label" ) then
                    element b { bdn-transform:transform($node/node(), $filename, $mode) }
                else
                    bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:choice )
        return
            bdn-transform:choice($node, $filename, $mode)
        case element (tei:div)
        return
            bdn-transform:div($node, $filename, $mode)
        case element ( tei:docImprint )
        return
            (bdn-transform:transform($node/node(), $filename, $mode),
            element br {})
        case element (tei:floatingText)
        return
            element div
            {
                attribute class { "floating-text" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:foreign )
        return
            element i
            {
                attribute lang { $node/@xml:lang },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:front )
        return
            element div
            {
                attribute class { "front" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:head )
        return
            bdn-transform:head($node, $filename, $mode)
        case element ( tei:hi )
        return
            bdn-transform:hi($node, $filename, $mode)
        case element ( tei:index )
        return ()
        case element ( tei:item )
        return
            element li
            {
                attribute class
                {
                    "list-item" ||
                    (if ( $node/ancestor::tei:div[@type = "contents"] ) then
                        "-toc"
                    else
                        ())
                },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:join )
        return
            switch($node/@result)
            case "p" return
                element div
                {
                attribute class { "p" },
                bdn-transform:apply-join($node, $filename, $mode)
                }
            case "item" return
                element li
                {
                attribute class { "list-item" },
                bdn-transform:apply-join($node, $filename, $mode)
                }
            (: the default case is that join has no @result :)
            default return
                bdn-transform:apply-join($node, $filename, $mode)
        case element ( tei:label )
        return
            if ( $node/@type = "headword" ) then
                element b
                {
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
            else
            (: test: the old transformations produced a span with .label-wrapper.
            There is no corresponding style in the css, so I don't think it is
            really needed :)
                bdn-transform:transform($node/node(), $filename, $mode)
        case element ( tei:l )
        return
            element span
            {
                attribute class { "line" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:lb )
        return
            (: test: preceding-sibling::*[1] cf. digital & critical :)
            if ($node/preceding-sibling::*[1]
                [ self::tei:titlePart
                or self::tei:docImprint
                or self::tei:head
                ]
            ) then
                ()
            else
                element br {}
        case element ( tei:lem )
        return
            bdn-transform:lem($node, $filename, $mode)
        case element ( tei:lg )
        return
            element div
            {
                attribute class { "linegroup" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:line )
        return
            element span
            {
                attribute class { "line" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:list )
        return
            element ul
            {
                attribute class
                {
                    if ( $node/parent::tei:item ) then
                        "nested-list"
                    else if ( $node/ancestor::tei:rdg ) then
                        "rdg-list"
                    else
                        "text-list"
                },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:milestone )
        return
            bdn-transform:milestone($node, $filename, $mode)
        case element ( tei:note )
        return
            bdn-transform:note($node, $filename, $mode)
        case element ( tei:opener )
        return
            if ( $node/@rend = "center-aligned" ) then
                element div
                {
                    attribute class { "center-aligned opener" },
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
            else
                ()
        case element ( tei:p )
        return
        (: since we have div nested in p, e.g. for compositor's
        errors, we need to choose div here instead of p.
        otherwise the HTML content model is violated. :)
        element div
        {
            attribute class
            {
                if ( $node/@rend= "center-aligned" ) then
                    "center-aligned"
                else
                    "p"
            },
            (: test: I guess this is not necessarily needed considering the provisions of the wrapper function :)
            if ($node/@copyOf) then
                bdn-transform:apply-corresp-node($node/@copyOf, $filename, $mode, "")
            else
                bdn-transform:transform($node/node(), $filename, $mode)
        }
        case element ( tei:pb )
        return
            if ( matches($node/@edRef, $mode) or $mode = "digital" or $mode = "critical" ) then
                (if ( $node/@break-before = "yes" ) then
                    text { " " }
                else
                    (),
                element span
                {
                    attribute class { "pb" },
                    "|" || replace( $node/@edRef, '[# ]+', '' )
                    || (if ( $node/@type = 'sp' ) then
                        "[" || $node/@n || "]"
                    else
                        $node/@n)
                    || "|"
                })
            else
                ()
        case element ( tei:persName )
        return
            if ( $mode = "digital" ) then
                (bdn-transform:transform($node/node(), $filename, $mode),
                if ( $node/@ref ) then
                    let $id := util:hash(generate-id($node), "md5")
                    let $ref := "#" || $id
                    let $uri := replace($node/@ref, "^#", "")
                    return
                    (
                        text{ " " },
                        element span
                        {
                            attribute class { "fa-user-link" },
                            attribute id { $id },
                            attribute data-target { $ref },
                            attribute tgURI { $uri },
                            element i
                            {
                                attribute class { "fas fa-user" }
                            }
                        }
                    )
                else
                    ())
            else
                bdn-transform:transform($node/node(), $filename, $mode)
        case element ( tei:ptr )
        return
            if ( $mode = "digital" ) then
                let $id :=
                    if ($node/@id) then
                        util:hash($node/@id, "md5")
                    else
                        util:hash(generate-id($node), "md5")
                let $ref := "#" || $id
                return
                (
                    text { " " },
                    if ( $node/@type = "editorial-commentary" ) then
                    (
                        bdn-transform:transform($node/node(), $filename, $mode),
                        element span
                        {
                            attribute class { $node/@type },
                            attribute href { $node/@target },
                            attribute data-toggle { "modal" },
                            <i class="far fa-comment"></i>
                        }
                    )
                    else if ( $node/@type = "page-ref" ) then
                    (: these are references in editorial commentaries that point to another editorial commentary
                     : problem: the element with the targeted id is not the link in the referenced document but the modal itself...
                     : remediation of this problem is saved for [issue 321](https://gitlab.gwdg.de/bibliothek-der-neologie/bdn-SADE/-/issues/321)
                     :)
                        element a
                        {
                            attribute class { "text-link" },
                            attribute href { bdn-transform:get-href-from-id(replace($node/@target, "#", ""), $mode) },
                            <i class="fas fa-external-link-alt"></i>
                        }
                    (: else encompasses $node/@type = "bibliographic-object" :)
                    else
                        element span
                        {
                            attribute class { "fa-book-link" },
                            attribute id { $id },
                            attribute data-target { $ref },
                            attribute tgURI { $node/@target },
                            <i class="fas fa-book"></i>
                        },
                        bdn-transform:make-modal($node, $mode, $id, $filename)
                )
            else
                ()
        case element ( tei:quote )
        return
            if ( $mode = "digital" ) then
                (element span
                {
                    attribute class { "quote" },
                    attribute id { util:hash(generate-id($node), "md5") },
                    attribute target-node { util:hash(generate-id(text {$node/text()}), "md5") },
                    bdn-transform:transform($node/node(), $filename, $mode),
                    element span
                    {
                        attribute class { "quote-icon" },
                        <i class="fas fa-quote-right"></i>
                    }
                },
                bdn-transform:make-quote($node, $mode, $filename))
            else
                ()
        case element ( tei:ref )
        return
        (
            element span
            {
                attribute class { "ref" },
                bdn-transform:transform($node/node(), $filename, $mode),
                string(" "),
                (: allow for multiple refs although not covered by the schema :)
                for $href in bdn-transform:get-href-from-id(replace($node/@target, "#", ""), $mode)
                return
                    element a
                    {
                        attribute class { "text-link" },
                        attribute href { $href },
                        <i class="fas fa-external-link-alt"></i>
                    }
            },
            if (bdn-transform:starts-next-text-with-punctuation($node, $mode)) then
                ()
            else
                text { " " }
        )
        case element ( tei:rdg )
        return
            if ( $mode = "critical" ) then
                if ( $node[@type = ("ppl", "ptl")] ) then
                    element div
                    {
                        attribute class { "rdg-crit" },
                        bdn-transform:transform($node/node(), $filename, $mode)
                    }
                else if ( $node[@type = ("pt", "pp")]
                    and not($node/preceding-sibling::tei:rdg/@type = $node/@type))
                then
                    bdn-transform:make-link($node, $filename, $mode)
                else
                    ()
            else
                bdn-transform:transform($node/node(), $filename, $mode)
                (: todo: braucht make-marker den zweiten Parameter? :)
        case element ( tei:rdgMarker )
        return
            bdn-transform:rdgMarker($node, $filename, $mode)
        case element ( tei:row )
        return
            if ( $node/preceding-sibling::tei:row ) then
                element table
                {
                    element tr
                    {
                        bdn-transform:transform($node/node(), $filename, $mode)
                    }
                }
            else
                element tr
                {
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
        case element ( tei:seg )
        return
            if ( $node/@type = "margin" ) then
                element span
                {
                    attribute class { "margin" },
                    text { "{" },
                    bdn-transform:transform($node/node(), $filename, $mode),
                    text { "}" }
                }
            else
                bdn-transform:transform($node/node(), $filename, $mode)
        case element ( tei:signed )
        return
            element div
            {
                attribute class { "signed" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:supplied )
        return
            (
                text { "[" || bdn-transform:transform($node/node(), $filename, $mode) || "]" }
            )
        case element ( tei:teiHeader )
        return ()
        case element ( tei:text )
        return
            if ( $mode = "digital" ) then
                element div
                {
                    attribute class { "col-md-11" },
                    bdn-transform:transform( $node/node(), $filename, $mode)
                }
            else
                bdn-transform:transform( $node/node(), $filename, $mode)
        case element ( tei:titlePage )
        return
            element span
            {
                attribute class { "title-page" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case element ( tei:titlePart )
        return
            bdn-transform:titlePart($node, $filename, $mode)
        default
        return
            bdn-transform:transform($node/node(), $filename, $mode)
};

(:~
 : Creates a whitespace depending on the flag @break-after that is set while
 : serializing the intermediate format.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node The current node
 : @return A whitespace if the flag is set
 :)
declare function bdn-transform:make-whitespace-after($node as node())
as text()*
{
    if ($node/@break-after = "yes") then
        if ($node/following::text()[1][starts-with(., ",") or starts-with(., ")")]) then
            ()
        else if ($node[self::tei:ptr]
        and $node/preceding::text()[1][ends-with(., "{") or ends-with(., "(")]) then
            ()
        else
            text {" "}
    else
        ()
};

(:~
 : Looks up the author of the work a given file belongs to. For works that have
 : already been published, no look up is necessary, but since links should also
 : work for unpublished documents to enable correcting links, we have created
 : a map, whitelist.xml, that contains all necessary information.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file The current file, e.g. "2228v"
 : @return The author of the work the file belongs to
 : @deprecated
 :)
declare function bdn-transform:get-author($file as xs:string*)
as xs:string*
{
    let $first-char := substring($file, 1, 1)
    let $uri :=
        (: assumes, tguris always start with a number...!?
         : => that is not true; deprecate and use bdn-utils:get-author-from-whitelist
         :)
        if (matches($first-char, "[0-9]")) then
            substring-before($file, ".")
        else
            ()
    return
        if ($uri) then
            whitelist:get-doc()//bdn:author[matches(following-sibling::bdn:files//*, $uri)]/string()
        else
            substring-before($file, "-")
};

(:~
: Retrieves the target node of a given @copyOf-reference. Since references can
: refer to another chapter (or even another volume) of a BdN work, the target
: is looked up in the ID file.
:
: @author Michelle Weidling
: @version 1.0
: @param $copyOf A reference to another node
: @param $filename The current file
: @return The node the @copyOf refers to
:)
declare function bdn-transform:find-corresp-node
    (
        $copyOf as xs:string,
        $filename as xs:string
    )
as node()?
{
    let $author := bdn-transform:get-author($filename)
    let $copyOf := replace($copyOf, "#", "")
    let $target-file := map:get($bdn-transform:id-map, $copyOf)
    let $doc := doc($config:edition || $author || "/xml/" || $target-file || ".xml")
    return
        $doc//*[@xml:id = $copyOf]
};

(:~
 : This function copies a TEI-node.
 :
 : @param $node A TEI-node with an attribute copyOf referencing another node
 : @return The desired copy as a constructed node
 :)
declare function bdn-transform:copy-node($node, $filename, $mode)
as node()*
{
    let $attributes := $node/@*[local-name(.) != "copyOf"]
    let $attributes :=
        for $attribute in $attributes return
            attribute {local-name($attribute)} {$attribute}
    let $result :=
        element { node-name($node) } {
            $attributes,
            bdn-transform:find-corresp-node($node/@copyOf, $filename)
        }
    return
        $result
};

(:~
: Finds and applies the node that is referenced with @copyOf.
: In case of tei:join we have to apply only the child nodes of the corresponding
: nodes because tei:join defines a new resulting element in @result that makes
: the original nodes obsolete.
:
: @author Michelle Weidling
: @version 1.0
: @param $copyOf A reference to another node
: @param $mode The current edition
: @param $file
: @param $scope The value of tei:join/@scope
: @return The transformed node the @copyOf refers to
:)
declare function bdn-transform:apply-corresp-node
    (
        $copyOf as xs:string,
        $filename as xs:string,
        $mode as xs:string,
        $scope as xs:string
    )
as node()*
{
    let $corresp-node := bdn-transform:find-corresp-node($copyOf, $filename)
    return
        if ( $scope = "branches" ) then
            for $child in $corresp-node return
                bdn-transform:transform($child, $filename, $mode)
        else
                bdn-transform:transform($corresp-node, $filename, $mode)
};

(:~
: In cases like tei:join several references are stated in a @target. This
: function returns all corresponding nodes.
:
: @author Michelle Weidling
: @version 1.0
: @param $target The @target of a e.g. tei:join
: @param $mode The current edition
: @return The nodes the @target refers to
:)
declare function bdn-transform:apply-join
    (
        $node as element (tei:join),
        $filename as xs:string,
        $mode as xs:string
    )
as node()*
{
    for $copyOf in tokenize($node/@target, " ")
        let $applied-node := bdn-transform:apply-corresp-node($copyOf, $filename, $mode, $node/@scope)
        let $first-text := $applied-node[1]/descendant-or-self::text()[1]
        let $conjunction :=
            if ( starts-with($first-text, ",") ) then
                text{ "" }
            else
                text{ " " }
        return ($conjunction, $applied-node)
};

declare function bdn-transform:is-next-text-in-app($node as node()*)
as xs:boolean
{
    if ( $node/following::text()[1][normalize-space(.) != ""][ancestor::tei:app] ) then
        true()
    else
        false()
};

declare function bdn-transform:is-next-app-omitted($node as node()*, $mode as xs:string)
as xs:boolean
{
    let $next-app := $node/following-sibling::*[1][self::tei:app]
    return
        if ( $next-app/child::tei:rdg[matches(@wit, $mode) and @type = "om"] ) then
        true()
        else
        false()
};

(:~
: Creates a random ID of 40 letters consisting of Basic Latin letters and
: numbers. This function is needed since the usage of generate-id() (which we
: originally used for the linking between icons and a modal dialogue) lead to
: ID duplettes.
:)
declare function bdn-transform:make-random-id()
as xs:string
{
    let $charset := "ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvwxyz0123456789"
    let $id :=
        for $iii in 1 to 40 return
            let $number := util:random(40)
            return
                substring($charset, $number, 1)
    return string-join($id, "")
};

(:~
 : Creates a span for a tei:rdg which is relevant for textual criticism
 : and which is not handled in the main text, i.e. tei:rdg[@type = ("v", "pp", "pt")].
 : All spans of a tei:div[@type = "section"] are displayed en bloc at the
 : end of the respective div.
 :
 : This function is called by critical:make-apparatus, critical:make-apparatus-entry and critical:make-link.
 :
 : @author Michelle Weidling
 : @author Stefan Hynek
 : @version 1.0
 : @param $node the current tei:rdg
 : @return element that is the ancestor of $node
 :)

declare function bdn-transform:get-rdg-ancestor($node as element(tei:rdg))
as element(*)
{
    if ($node/ancestor::tei:app[@type = "standalone"][not(descendant::tei:div[@type])]) then
        $node/ancestor::tei:app[@type = "standalone"]
    else if ($node/ancestor::tei:head[not(ancestor::tei:div[@type = "section"])]) then
        $node/ancestor::tei:head
    else
        $node/ancestor::tei:div[@type][1]
};

(:~
 : Creates a modal dialog for an abbreviation expansion or the display of an
 : editorial commentary. Since both use cases nearly have the same requirements
 : both are handled by this function.
 :
 : @author Michelle Weidling
 : @author Simon Sendler
 : @version 1.1
 : @param $node a tei:abbr or a tei:note[@type = "editorial-commentary"]
 : @param $id
 : @param $filename e.g. "2228v-if.xml"
 : @return div the modal dialog
 :)
declare function bdn-transform:make-modal
    (
        $node as node(),
        $mode as xs:string,
        $id as xs:string,
        $filename as xs:string*
    )
as element(div)
{
    let $title :=
        if ( $node[self::tei:abbr] ) then
            let $expans :=
                for $expan in $node/following-sibling::tei:expan
                return
                (
                    bdn-transform:transform($expan, $filename, $mode),
                    if (not($expan = $node/following-sibling::tei:expan[last()])
                        and count($node/following-sibling::tei:expan) gt 1)
                    then
                        ", "
                    else
                        ()
                )
            let $expans := string-join($expans)
            (: the abbr node can contain another entity
            : -> its text() would be empty and is contained
            : in a child node
            :)
            let $abbr :=
                if ($node[self::tei:abbr]/text()) then
                    $node[self::tei:abbr]/text()
                else
                    $node[self::tei:abbr]/*/text()
            let $child-nodes :=
                for $child-node in $abbr
                return
                    bdn-transform:transform($child-node, $filename, $mode)
            let $child-nodes := string-join($child-nodes)
            return
                "Abkürzungsauflösung von """ || $child-nodes || """: " || $expans
        else if ( $node/@type = "editorial-commentary" ) then
            bdn-transform:transform($node/tei:label, $filename, $mode)
        else if ( $node/@type = "editorial-footnote" ) then
            bdn-transform:transform($node/node(), $filename, $mode)
        else if ( $node/tei:corr )
        then
            let $corr :=
                if ( $node/tei:sic[not(child::node())] )
                then
                    element em { "∅" }
                else
                    bdn-transform:transform($node/tei:sic, $filename, $mode)
            return
                ( if ( $node/tei:corr/@type = "authorial" ) then
                    "Autorschaftliche"
                  else
                    "Editorische"
                )
                (: $corr can be a multi-node sequence, so the text of these nodes must be joined :)
                || " Korrektur von: " || string-join($corr) || " (" || $mode || ")"
        else
            ()
    let $content :=
        if ( $node/@type = "editorial-commentary" ) then
            bdn-transform:transform($node/tei:p, $filename, $mode)
        else
            ()
    let $class :=
        if ( $node/tei:corr ) then
            "modal-typo"
        else
            "abbr-or-commentary-modal"
    return
        element div
        {
            attribute class { "modal fade" },
            attribute role { "dialog" },
            attribute id { $id },
            element div
            {
                attribute class { "modal-dialog " || $class },
                element div
                {
                    attribute class { "modal-content" },
                    element div
                    {
                        attribute class { "modal-header" },
                        element button
                        {
                            attribute type { "button" },
                            attribute class { "close" },
                            attribute data-dismiss { "modal" },
                            text { "x" }
                        },
                        if ( $node/tei:corr ) then
                            $title
                        else
                            element h5
                                {
                                    attribute class { "modal-title abbr-or-commentary-head" },
                                    (: following might go? :)
                                    if ( $node[self::tei:abbr or self::tei:choice] ) then
                                        ()
                                    else
                                        (element i
                                        {
                                            attribute class { "far fa-comment" }
                                        },
                                        text { " " }),
                                    $title
                                }
                    },
                    if ( $content ) then
                        element div
                        {
                            attribute class { "modal-body" },
                            element table
                            {
                                element tr
                                {
                                    element td { $content }
                                }
                            }
                        }
                    else
                        ()
                }
            }
        }
};

(:~
: Creates the icon for a compositor's error and initiates the construction of
: the respective modal div.
:
: This function is a merger of the earlier, separate functions handling
: typo corrections for the different views.
:
: @author Michelle Weidling
: @author Simon Sendler
: @version 2.0
: @param $node The the current tei:choice
: @param $mode The current edition
: @return The modal div, a whitespace and the icon
:)
declare function bdn-transform:make-typo-corr
    (
        $node as element (tei:choice),
        $filename as xs:string,
        $mode as xs:string
    )
as node()*
{
    let $random-id := bdn-transform:make-random-id()
    let $id := util:hash($random-id, "md5")
    return
        (bdn-transform:transform($node/tei:corr, $filename, $mode),
        bdn-transform:make-modal($node, $mode, $id, $filename),
        element a
        {
            attribute class { "comp-error" },
            attribute data-target { "#" || $id },
            element i
            {
                attribute class { "fa fa-wrench" },
                attribute aria-hidden { "true" },
                attribute data-toggle { "modal" }
            }
        })
};

(:~
 : Triggers the creation of a modal dialog for compositor's errors in different
 : printings and the base text which have have been corrected either by the
 : author himself or by modern editors.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node a tei:app which has a child with a compositor's error
 : @param $filename the filename, e.g. "2228v.xml"
 : @return the text to be displayed and the (hidden) modal dialog
 :)
declare function bdn-transform:make-typo-corrs($node as element(tei:app), $filename as xs:string,
    $mode as xs:string)
as node()*
{
    let $typo-corr-rdgs := $node/child::tei:rdg[@type = "typo-correction"][descendant::tei:choice]
    let $typo-corr-lem := $node/child::tei:lem[descendant::tei:corr]
    let $wit :=
        if ( $typo-corr-rdgs ) then
            ()
        else
            (: single file names don't have a hyphen, only the ones that are the
            result of semantic splitting :)
            if ( matches($filename, "-") ) then
                let $author := substring-before($filename, "-")
                let $full-file := doc($config:full-xmls || $author || "-full.xml")
                return $full-file//tei:witness[@n = "base-text"]/@xml:id/string()
            else
                $node/ancestor::tei:TEI//tei:witness[@n = "base-text"]/@xml:id
    return
        if ( $typo-corr-rdgs ) then
            for $typo in $typo-corr-rdgs
            return
                try {
                    (: (bdn-transform:transform($typo//tei:sic, $filename, $mode), :)
                    bdn-transform:make-typo-corr($typo//tei:choice[child::tei:corr], $filename, replace($typo/@wit, "[# ]", ""))
                    (: ) :)
                } catch * {
                    bdn-transform:error-message($err:code, $err:description, $typo//tei:choice, "bdn-transform:make-typo-corrs", $filename)
                }
        else
        (
            try {
                (: bdn-transform:transform($typo-corr-lem//tei:choice, $filename, $mode), :)
                bdn-transform:make-typo-corr($typo-corr-lem//tei:choice[child::tei:corr], $filename, $wit)
            } catch * {
                bdn-transform:error-message($err:code, $err:description, $typo-corr-lem//tei:choice[child::tei:corr],
                   "bdn-transform:make-typo-corrs", $filename)
            }
        )
};

(: ~
 : Inserts a span containing relevant information on an error into the edition
 : text for improved visibility of errors during transformation or in the markup.
 :
 : Codes & descriptions according to http://www.w3.org/2005/xqt-errors
 :
 : @author Simon Sendler
 : @version 1.0
 : @param $code An XQuery error code
 : @param $description An error's description
 : @param $node An XML-element
 : @param $function The name of the function during the execution of which the error was raised
 : @param $filename The name of the file being transformed
 :)
declare function bdn-transform:error-message($code as xs:string, $description as xs:string,
    $node as node(), $function as xs:string, $filename as xs:string)
as element(div)
{
    util:log( "debug",
        "Error during transformation of " || $filename || ": " || $code || $description || $node/name()),
    element div
    {
        attribute class { "error" },
        text { "Error in function " || $function || ": " || $code },
        element br {},
        text { "Description: " || $description },
        element br {},
        text { "Offending node: " || $node/name() }
    }
};

(:~
 : Checks if the next text node of the main text begins with punctuation.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node The current node
 : @param $mode The current view. In case of the single edition view, $mode
 : denotes the current printing
 : @return true() if next text node begins with punctuation
 :)
declare function bdn-transform:starts-next-text-with-punctuation($node as node(),
$mode as xs:string) as xs:boolean
{
  let $next-valid-text :=
    if ($mode = ("app", "digital")) then
      $node/following::text()[not(normalize-space(.) = "")
      (: this is important for cases where we have more than one rdg of a tei:app.
      without this query the next valid text node would be the text node of the
      following tei:rdg, whereas we want the next main text :)
      and not(./ancestor::tei:rdg[@type = ("v", "pp", "pt")])][1]
    else
      $node/following::text()[not(normalize-space(.) = "")
      (: in the single edition view, we don't mind text nodes that appear
      in tei:rdgs other than the current printing :)
      and not(./ancestor::tei:rdg[not(matches(./@wit, $mode))])][1]

  let $first-char := substring($next-valid-text, 1, 1)
  return
    if (matches($first-char, "[–\?!\.,\)};“:]")) then
      true()
    else
      false()
};

(:~
 : Creates a relative URL to an object given in a tei:ref's target or tei:choice's
 : resp attribute.
 : In the BdN, a tei:ref denotes a reference to another chapter, paragraph, note,
 : etc. within the current BdN work. A tei:choice with an @corresp is part of an
 : authorial corrigendum.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node A tei:ref or tei:choice element
 : @return The complete relative URL to the referenced object, e.g.
 : "book.html?file=griesbach-1-2&edition=digital#gr_section_19"
 :)
declare function bdn-transform:get-link-url($node as element(*), $mode as xs:string)
as xs:string
{
  let $id :=
    if ($node[self::tei:ref]) then
      replace($node/@target, "#", "")
    (: $node[self::tei:choice] :)
    else
      replace($node/@corresp, "#", "")
  let $target-file := id-file:get-id-file()//*[./string() = $id]/parent::bdn:file/@uri/string()

  return "book.html?file=" || $target-file || "&amp;edition=" || $mode || "#" || $id
};

(:~
 : Create a URL to an anchor from a given ID. This uses ID lookup with the help of an ID-file.
 : This will later be replaced with an index lookup, see https://gitlab.gwdg.de/bibliothek-der-neologie/bdn-SADE/-/issues/305
 :)
declare function bdn-transform:get-href-from-id($ids as xs:string+, $mode as xs:string)
as xs:string+
{
    for $id in $ids
    return "book.html?file=" || id-file:get-id-file()//*[./string() = $id]/parent::bdn:file/@uri/string() || "&amp;edition=" || $mode || "#" || $id
};

declare function bdn-transform:get-link-url-to-corresp-node($node as element(*), $filename as xs:string,
$mode as xs:string) as xs:string
{
    "book.html?file=" || $filename || "&amp;edition=" || $mode || "#" || $node/@xml:id
};

(:~
 : Retrieves the color that belongs to one or several witnesses of a tei:rdg.
 :
 : This function is called by bdn-transform:transform.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $wits tei:rdg/@wit
 : @return the hex color codes a witness
 :)
declare function bdn-transform:get-rdg-color($wits as xs:string*) as xs:string
{
    if (count(tokenize($wits, " ")) gt 1) then
        let $wit-combination :=
            tokenize($wits, " ")
            => string-join("")
        let $color-hex := map:get($bdn-transform:colors, $wit-combination)
        return
            if ($color-hex) then
                $color-hex
            else
                map:get($bdn-transform:colors, "other")
    else
      map:get($bdn-transform:colors, $wits)
};

(:~
 : Creates an span for a tei:rdg that is relevant for textual criticism,
 : i.e. a tei:rdg[@type = ("om", "v", "pp", "pt", "ppl", "ptl")], and an
 : path that belongs to it.
 :
 : While the span will be displayed in the div[@id = "variants"] on the
 : page, the path will be part of the SVG layer. Moving the elements that
 : are created in this function to the right place is handled by xml-utils:make-html-digital-view.
 :
 : This function is called by bdn-transform:transform.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node a tei:rdg
 : @param $filename e.g. "2228v-if.xml"
 : @return sequence of span and path
 :)
declare function bdn-transform:make-rdg($node as element(tei:rdg), $mode as xs:string, $filename as xs:string)
as element()*
{
  let $wits := replace($node/@wit, "[#]+", "")
  let $color := bdn-transform:get-rdg-color($wits)

  let $wits := replace($wits, " ", "")
  let $rdg-id := util:hash(generate-id($node), "md5")
  let $lem-id := util:hash(generate-id($node/parent::*), "md5")
  let $target-lem := "#" || $lem-id
  let $target-rdg := "#" || $rdg-id
  return
  (
    element span
    {
      attribute class {"rdg-digital rdg-" || $wits},
      attribute id {$rdg-id},
      attribute target-lem {$target-lem},
      attribute wits {$wits},
      <table>
        <tr>
          <td class="siglum">
            <strong>{$wits}:</strong>
          </td>
          <td>
            {if ($node/@type = "om") then
              <i>&#248;</i>
            else
              if ( $node/ancestor::tei:label[@type = "headword"]) then
                element b
                {
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
            else
                bdn-transform:transform($node/node(), $filename, $mode)
            }
          </td>
        </tr>
      </table>
    },
    element path
    {
      attribute class {"rdg-path"},
      attribute d {"M10 10"},
      attribute target-rdg {$target-rdg},
      attribute rdg-wits {$wits},
      attribute target-lem {$target-lem},
      attribute stroke {$color},
      attribute fill {"transparent"}
    }
  )
};

(:~
 : Creates a link in the main text to the critical apparatus at the end of a div
 : for a given variant or insertion.
 :
 : This function is called by critical:transformation.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 : @version 1.0
 : @param $node the current tei:rdg
 : @return an span with the tei:rdg's witness and the wit's text
 :)
declare function bdn-transform:make-link($node as element(tei:rdg), $filename as xs:string,
$mode as xs:string)
as element(span)
{
  let $ancestor := bdn-transform:get-rdg-ancestor($node)

  let $number := count($node/preceding::tei:rdg[ancestor::tei:div = $ancestor][@type = ("v", "pp", "pt")]) + 1
  let $type := $node/@type
  let $wits := string-join($node/parent::tei:app/tei:rdg[@type = $type]/@wit, "")
  let $wits := replace($wits, "[# ]+", "")
  let $id := util:hash($node/@id, "md5") || "-" || $number

  return
    element span
    {
      attribute class {"insertion-wrapper"},
      (if ($node[@type = ("pt", "pp")]) then
        text {" ∥"}
      else
        ()),
      element a
      {
        attribute href {"#" || $id},
        <sup>{$wits}{$number}</sup>
      }
    }
};

(:~
 : Creates the scribal abbreviations.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $wits The sigla for one or more printing(s)
 : @param $slash A Boolean parameter indicating whether the mark needs a slash
 : @param $mark A flag indicating whether the siglum is an opening or closing one
 : @return element(sub) The finished siglum
 :)
declare function bdn-transform:make-marker($wits as xs:string, $slash as xs:boolean,
$mark as xs:string) as element(sup)
{
  let $open :=
    if ($slash and $mark = "open") then
      "/"
    else
      ()
  let $close :=
    if ($slash and $mark = "close") then
      "\"
    else
      ()
  let $content := $open || $wits || $close

  return
    element sup
    {
      (: in cases or several printings these are divided by a whitespace :)
      replace($content, " ", "")
    }
};

declare function bdn-transform:make-quote($node as element(tei:quote), $mode as xs:string,
$filename as xs:string) as element()*
{
    let $text-content := text {$node/text()}
    let $id := util:hash(generate-id($text-content), "md5")
    let $corresp-ids := tokenize($node/@corresp, " ")
    let $corresp-ids :=
        for $corr-id in $corresp-ids return
            substring-after($corr-id, "#")

    let $no-of-quoted-texts := count($corresp-ids)

    let $target-id := "#" || util:hash(generate-id($node), "md5")

    let $corresp-nodes :=
        for $corr-id in $corresp-ids return
            (: in this case we are in a split file and $filename contains the author's name :)
            if (contains($filename, "-")) then
                let $author :=
                    if (contains($filename, "bahrdt-semler")) then
                        "bahrdt-semler"
                    else
                        substring-before($filename, "-")
                let $full-file := doc($config:full-xmls || $author || "-full.xml")
                return
                    $full-file//*[@xml:id = $corr-id]

            (: in this case we are in a single file and $filename is the TextGrid URI :)
            else
                for $doc in collection($config:sade-data) return
                    if ($doc//*[@xml:id = $corr-id]
                    and not(matches(base-uri($doc), "full"))) then
                        $doc//*[@xml:id = $corr-id]
                    else
                        ()
    let $target-urls :=
        (: in this case we are in a split file and all IDs are listed in the ID file :)
        if (contains($filename, "-")) then
            bdn-transform:get-href-from-id($corresp-ids, "digital")

        (: in this case we are in a single file and ID are not listed in the ID file :)
        else
        for $corr-node in $corresp-nodes return
            for $doc in collection($config:sade-data) return
                if ($doc//* = $corr-node
                and not(matches(base-uri($doc), "full"))) then
                    bdn-transform:get-link-url-to-corresp-node($corr-node, base-uri($doc), "digital")
                else
                    ()

    let $quoted-texts :=
        for $corr-node in $corresp-nodes return
            $corr-node/ancestor::tei:text//tei:supplied[@reason = "toc-title"][1]

    return
        (
            element div
            {
                attribute class {"quoted-text"},
                attribute id {$id},
                attribute target-node {$target-id},
                for $iii in 1 to $no-of-quoted-texts return
                    <table class="quoted-passage" count="{$iii}">
                        <tr>
                            <td class="siglum">
                                Zitat aus <span class="quoted-work">{$quoted-texts[$iii]/string()}</span>:
                            </td>
                        </tr>
                        <tr>
                            <td class="direct-quote">
                                {
                                    bdn-transform:transform($corresp-nodes[$iii]/node(), $filename, $mode)
                                }
                            </td>
                        </tr>
                        <tr>
                            <td class="quote-link">
                                <a href="{$target-urls[$iii]}">
                                    Gehe zum zitierten Text
                                    <i class="fas fa-external-link-alt"></i>
                                </a>
                            </td>
                        </tr>
                    </table>
            },
            element path
            {
                attribute class {"quote-path"},
                attribute d {"M10 10"},
                attribute target-quoting {$target-id},
                attribute target-quoted {"#" || $id},
                attribute stroke {map:get($bdn-transform:colors, "quote")},
                attribute fill {"transparent"}
            }
        )
};

declare function bdn-transform:crit-app-entry($rdg as element(tei:rdg)*,
$number as xs:integer, $filename as xs:string)
as element(span)
{
    let $wits :=
        string-join($rdg/../tei:rdg[@type = ("v", "pp", "pt")]/@wit, "")
        => replace("[# ]+", "")
    let $id := util:hash($rdg/@id,"md5") || "-" || $number
    return
        element span
        {
            attribute class { "app-entry" },
            attribute id { $id },
            element b
            {
                $wits || $number || ": "
            },
            if ( $rdg/ancestor::tei:label[@type = "headword"]) then
                element b
                {
                    bdn-transform:transform($rdg/node(), $filename, "critical")
                }
            else
                bdn-transform:transform($rdg/node(), $filename, "critical"),
            if ( $rdg/following-sibling::tei:rdg[@type = ("v", "pt", "pp" )] ) then
                (" (" || $rdg/@wit => replace("#", "") || ")",
                (for $following in $rdg/following-sibling::tei:rdg[@type = ("v", "pt", "pp" )] return
                    string-join(("; ",
                    if ( $rdg/ancestor::tei:label[@type = "headword"]) then
                        element b
                        {
                            bdn-transform:transform($rdg/node(), $filename, "critical")
                        }
                    else
                        bdn-transform:transform($rdg/node(), $filename, "critical"),
                    " (", $following/@wit => replace("#", ""), ")" ), ""))
                )
            else
                ()
        }
};

declare function bdn-transform:crit-app($node as node()*, $filename as xs:string,
$mode as xs:string)
as element(div)
{
    let $apps := $node//tei:app[tei:rdg[@type = ("v", "pt", "pp")]]
    (: only consider the first reading of each tei:app so that only one entry is
       created per tei:app :)
    let $rdgs := $apps//tei:rdg[@type = ("v", "pt", "pp")]
    return
        element div
        {
            attribute class { "critical-apparatus" },
            for $rdg in $rdgs
                let $ancestor := bdn-transform:get-rdg-ancestor($rdg)
                let $number := count($rdg/preceding::tei:rdg[ancestor::tei:div = $ancestor][@type= ("v", "pt", "pp")]) + 1
                return
                    bdn-transform:crit-app-entry($rdg, $number, $filename)
        }
};

declare function bdn-transform:apparatus($node, $filename, $mode) {
    switch( $mode )
        case "digital"
        return
            if ( $node/@type = "structural-variance" ) then
                bdn-transform:transform($node/tei:lem, $filename, $mode)
            else
                let $textcritical-rdgs := $node/child::*[@type = ("om", "v", "pp", "pt", "ppl", "ptl")]
                let $textcritical-no := count($textcritical-rdgs)
                let $typo-corrs := $node/child::*[@type = ("typo-correction")]
                return
                    (
                    if ( $typo-corrs ) then
                        bdn-transform:make-typo-corrs($node, $filename, $mode)
                    else
                        bdn-transform:transform($node/tei:lem, $filename, $mode),
                    for $rdg in $textcritical-rdgs return
                        (
                        bdn-transform:make-rdg($rdg, $mode, $filename),
                        if ($textcritical-no gt 1
                            and $rdg = $textcritical-rdgs[last()])
                        then
                            ()
                        else
                            element div
                            {
                                attribute class { "rdg-digital-divider" }
                            }
                        )
                    )
        case "critical"
        return
            if ($node/@type = "structural-variance") then
                if ($node/tei:lem[@type = "missing-structure"]) then
                    ()
                else if ($node/tei:rdg[@type = "varying-structure"]) then
                    bdn-transform:transform($node/tei:lem, $filename, $mode)
                else
                    ()
                else
                    (bdn-transform:transform($node/node(), $filename, $mode),
                    bdn-transform:make-whitespace-after($node))
        default
        return
            (if ($node/child::*[matches(@wit, $mode)]) then
                bdn-transform:transform($node/child::*[matches(@wit, $mode)], $filename, $mode)
            else
                bdn-transform:transform($node/child::tei:lem, $filename, $mode),
            if (bdn-transform:is-next-text-in-app($node)
            and bdn-transform:is-next-app-omitted($node, $mode)) then
                ()
            else
                bdn-transform:make-whitespace-after($node))
};

declare function bdn-transform:choice($node, $filename, $mode) {
    if ( $node/tei:abbr ) then
                bdn-transform:transform($node/tei:abbr, $filename, $mode)
            else if ( $mode = "critical" and $node/tei:corr ) then
                bdn-transform:transform($node/tei:orig, $filename, $mode)
            else if ( $node/tei:orig ) then
                bdn-transform:transform($node/tei:orig, $filename, $mode)
            else if ( $node/tei:corr[@type = "editorial"] ) then
                bdn-transform:make-typo-corr($node, $filename, $mode)
            else if ( $node/tei:corr[@type = "authorial"] ) then
                (bdn-transform:transform($node/tei:corr, $filename, $mode),
                text { " " },
                element a
                {
                    attribute href { bdn-transform:get-link-url($node, $mode) },
                    element i
                    {
                        attribute class { "fas fa-pen-fancy" }
                    }
                })
            else if ( $node/tei:corr
                    and ($node/..[@type = "typo-correction"][contains(@wit, $mode)]
                    or not($node/parent::tei:lem/following-sibling::tei:rdg[contains(@wit, $mode)])) ) then
                (bdn-transform:transform($node/tei:corr, $filename, $mode),
                bdn-transform:make-typo-corr($node, $filename, $mode),
                (: compositor's error with authorial correction :)
                if ($node/tei:corr[@type = "authorial"]) then
                    element a
                    {
                        attribute href {bdn-transform:get-link-url($node, $mode)},
                        <i class="fas fa-pen-fancy"></i>
                    }
                else
                    ()
                )
            else
                ()
};

declare function bdn-transform:div($node, $filename, $mode) {
    switch($node/@type)
        case "synopsis"
        return
            (
            bdn-transform:transform($node/tei:head, $filename, $mode),
            element table
            {
                attribute class { "synopsis" },
                element tr
                {
                    element td
                    {
                        bdn-transform:transform($node/tei:div[1]/tei:head, $filename, $mode)
                    },
                    element td
                    {
                        bdn-transform:transform($node/tei:div[2]/tei:head, $filename, $mode)
                    }
                },
                (: ATTENTION: this implies that the two div[@type = 'letter']
                always have the same number of ab :)
                (
                for $ab in $node/tei:div[1]/tei:ab
                return
                    element tr
                    {
                        element td
                        {
                            attribute class { "padding-s" },
                            bdn-transform:transform($ab/node(), $filename, $mode)
                        },
                        element td
                        {
                            attribute class { "padding-s" },
                            bdn-transform:transform($ab/ancestor::tei:div[@type = 'synopsis']//tei:ab[matches($ab/@corresp, @xml:id)], $filename, $mode)
                        }
                    }
                )
            }
            )
        case "letter"
        return
            (: if ($node/ancestor::tei:div[@type = "synopsis"]) then
                ()
            else :)
                bdn-transform:transform($node/node(), $filename, $mode)
        default
        return
            (element div
                {
                    attribute class {$node/@type},
                    bdn-transform:transform($node/node(), $filename, $mode)
                },
            if ( $node[@type = ("section", "contents", "preface", "corrigenda", "split", "entry")]
                and $mode = "critical" and not($node/ancestor::tei:rdg) ) then
                bdn-transform:crit-app($node, $filename, $mode)
            else
                ())
};

declare function bdn-transform:head($node, $filename, $mode)
{
    switch($node/@type)
        case "main"
        return
            element div
            {
                attribute class { "list-main-head" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        case "sub"
        return
            element div
            {
                attribute class { "list-sub-head" },
                bdn-transform:transform($node/node(), $filename, $mode)
            }
        default
        return
            element h2
            {
            if ( $node/@copyOf and $mode = "single" ) then
                bdn-transform:apply-corresp-node($node/@copy-of, $filename, $mode, "")
            (: else if ( $mode = "critical"
                and $node[descendant::tei:rdg[@type = ("pp", "pt", "v")]
                and not(ancestor::tei:div[@type = "section"])])
            then
                bdn-transform:crit-app($node, $filename, $mode) :)
            else
                bdn-transform:transform($node/node(), $filename, $mode)
            }
};

declare function bdn-transform:hi($node, $filename, $mode)
{
    element em
    {
        if ( $node/@rend ) then
            attribute class
            {
                $node/@rend ||
                (if ( $node/@rend = "spaced-out"
                    and matches(substring($node/following::text()[1], 1, 1), "[,!\?\.:;]")
                ) then
                    " spaced-out--ending"
                else
                    ())
            }
        else
            bdn-transform:transform($node/node(), $filename, $mode)
    }
};

declare function bdn-transform:lem($node, $filename, $mode)
{
    if ( $mode = "digital" ) then
        let $relevant-siblings := $node/following-sibling::*[@type = ("om", "v", "pp", "pt", "ppl", "ptl")]
        let $relevant-wits := string-join($relevant-siblings/@wit, "")
        let $relevant-wits := replace($relevant-wits, "[# ]+", "")
        let $id := util:hash(generate-id($node/parent::*), "md5")
        return
            if ( $node/@type = "missing-structure" ) then
                ()
            else if ($node/following-sibling::*[@type = ("pt", "ptl")]) then
                let $insertion-rdgs := $node/following-sibling::*[@type = ("pt", "ptl")]
                let $insertion-wits := string-join($insertion-rdgs/@wit, "")
                let $insertion-wits := replace($insertion-wits, "[# ]+", "")
                return
                    element span
                    {
                        attribute class { "lem insertion" },
                        attribute id { $id },
                        attribute wits { $insertion-wits },
                        element sup { $insertion-wits },
                        text { "√" }
                    }

            else if ( count($node/*) = 1 and $node/tei:item ) then
                element li
                {
                    (
                    if ( $node/ancestor::tei:div[@type = "contents"] ) then
                        attribute class { "list-item-toc" }
                    else
                        attribute class { "list-item" }
                    ),
                    element span
                    {
                        attribute class
                        {
                            "lem" ||
                            (
                                if ($node/tei:div
                                    or $node/tei:titlePage
                                    or $node/tei:p
                                    or $node/tei:note)
                                then
                                    " lem-block"
                                else
                                    ()
                            )
                        },
                        attribute id { $id },
                        if ( $relevant-wits ) then
                            attribute wits { $relevant-wits }
                        else
                            (),
                        bdn-transform:transform($node/tei:item/node(), $filename, $mode)
                    }
                }
            else
                element span
                {
                    attribute class
                    {
                        "lem" ||
                        (
                        if ($node/tei:div
                            or $node/tei:titlePage
                            or $node/tei:p
                            or $node/tei:note)
                        then
                            " lem-block"
                        else
                            ()
                        )
                    },
                    attribute id { $id },
                    if ( $relevant-wits ) then
                        attribute wits { $relevant-wits }
                    else
                        (),
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
    else
        bdn-transform:transform($node/node(), $filename, $mode)
};

declare function bdn-transform:milestone($node, $filename, $mode)
{
    let $edition := replace( replace( $node/@edRef, "#", "" ), " ", "" )
    return
    if ( $mode = ("digital", "critical") ) then
        if ( $node/@unit = "fn-break" ) then
            element span
            {
                attribute class { "fn-break" },
                string("|" || $edition || $node/@n || "|")
            }
        else
            let $edition := replace( $node/@edRef, "[# ]+", "" )
            let $ancestor-rdg-type := $node/ancestor::tei:rdg[1]/@type
            return
                if ( $ancestor-rdg-type = ("ppl", "ptl")
                    and ($node/preceding-sibling::*[1][not(self::tei:pb
                    or self::tei:list
                    or self::tei:rdgMarker)]
                    or $node/parent::tei:seg/preceding-sibling::node()) ) then
                (<br />,
                if ($node/@unit = "p") then
                    element span
                    {
                        attribute class { "filler" }
                    }
                else
                    ())
                else
                    element span
                    {
                        attribute class { "milestone" },
                        switch( $node/@unit )
                        case "p" return "⌇⌇"
                        case "no-p" return <s>⌇⌇</s>
                        case "line" return "⌇"
                        case "no-line" return <s>⌇</s>
                        default return "",
                        <sup>{ $edition }</sup>,
                        " "
                    }
    else
        if ( $node/@unit = "fn-break"
            and $edition = $mode) then
            element span
            {
                attribute class { "fn-break" },
                string("|" || $edition || $node/@n || "|")
            }
        else
            ()
};

declare function bdn-transform:note($node, $filename, $mode)
{
    if ( $mode = "digital" ) then
        if ( $node/@type = "editorial-footnote" ) then
            let $id := if ($node/@xml:id) then $node/@xml:id else util:hash(generate-id($node), "md5")
            let $target := "#" || $id
            return
            (
                element span
                {
                    attribute class { $node/@type },
                    attribute href { $target },
                    attribute data-toggle { "modal" },
                    <i class="far fa-comment"></i>
                },
            bdn-transform:make-modal($node, $mode, $id, $filename)
            )
        else if ( $node/@type = "editorial-commentary" ) then
            bdn-transform:make-modal($node, $mode, $node/@xml:id, $filename)
        else
            element div
            {
                attribute class { "note" },
                bdn-transform:transform($node/node(), $filename,  $mode)
            }
    else if ( $node/@type = "editorial-commentary" ) then
            ()
    else
        (if ( $node/ancestor::tei:p ) then
            element br {}
          else
            (),
        element div
        {
            attribute class { "note" },
            bdn-transform:transform($node/node(), $filename, $mode)
        })
};

declare function bdn-transform:rdgMarker($node, $filename, $mode)
{
    if ( $mode = "critical" ) then
        if ( $node/@type = ("ppl", "ptl" )
            and $node/@context = "rdg")
        then
        (
            bdn-transform:make-marker($node/@wit, false(), $node/@mark),
            if ( $node/following-sibling::*[1][self::tei:rdgMarker]
                and ($node/following-sibling::node()[1][self::tei:rdgMarker]
                or normalize-space($node/following-sibling::node()[1]) = ""))
            then
                text { " " }
            else
                ()
        )
        else if ( $node/@type = ("ppl", "pp" )
            and $node/@mark = ("open", "close")
            and $node/@context = "lem")
        then
            bdn-transform:make-marker($node/@wit, true(), $node/@mark)
        else if ( $node/@type = "om" )
        then
            bdn-transform:make-marker($node/@wit, true(), $node/@mark)
        else if ( $node/@type = "v"
            and $node/@context = "lem"
            and $node/@mark = "close")
        then
            for $ref in tokenize($node/@ref, " ")
            let $corresp := $node/ancestor::*[last()]//tei:rdg[@id = $ref]
            return
                if ( not($corresp/preceding-sibling::tei:rdg/@type = $node/@type) ) then
                bdn-transform:make-link($corresp, $filename, $mode)
                else ()
        else
            ()
    else
        ()
};

declare function bdn-transform:titlePart($node, $filename, $mode)
{
    if ( $mode = "critical"
        and $node/ancestor::tei:rdg[@type = ("ppl", "ptl", "v", "pt", "pp")] )
    then
        element span
        {
            attribute class {"rdg-head"},
            bdn-transform:transform($node/node(), $filename, $mode)
        }
    else
        switch( $node/@type )
            case "main"
            return
                element h1
                {
                    attribute class { "main-title" },
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
            case "sub"
            return
                element h2
                {
                    attribute class { "sub-title" },
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
            case "volume"
            return
                element div
                {
                    attribute class { "volume" },
                    bdn-transform:transform($node/node(), $filename, $mode)
                }
            default
            return ()
};